from .chord import (ROI_to_degree,
                    plot_chord,
                    plot_rim_and_labels,
                    get_character_degree_locations,
                    plot_rim,
                    plot_arcs,
                    plot_arc,
                    plot_ROI_circles,
                    polar_to_cart,
                    cart_to_polar)
from .convert import convert_matrix
from .coord_labeler import (get_idx_to_label,
                            find_closest,
                            AttrDict,
                            get_yeo_atlas)
from .glassbrain import plot_glassbrain
from .patch_RendererAgg import (NeuoChordRenderAgg,
                                do_monkey_patch)
from .combine import combine_imgs, plot_and_combine


import numpy as np
import matplotlib.pyplot as plt
from nilearn import plotting
import matplotlib
from typing import Union

from nichord.chord import _network_order_colors_check, _threshold_proc


def plot_glassbrain(idx_to_label: dict,
                    edges: Union[list, np.ndarray],
                    edge_weights: Union[None, list, np.ndarray],
                    fp_glass: Union[None, str],
                    coords: Union[list, np.ndarray] = None,
                    cmap: Union[None, str, matplotlib.colors.Colormap] = None,
                    node_size: Union[None, float, int, list, np.ndarray] = 5,
                    linewidths: Union[None, float, int, list] = 6,
                    alphas: Union[None, float, int, list] = None,
                    network_order: Union[None, list] = None,
                    network_colors: Union[None, dict] = None,
                    dpi: int = 400,
                    vmin: Union[None, int, float] = None,
                    vmax: Union[None, int, float] = None,
                    edge_threshold: Union[float, int, str] = 0.,
                    ) -> None:
    """
    Plots the connectome on a brain, where the node colors correspond to the
        ROI's label (specified by idx_to_label). Each edge's color correspond
        to the edge's weight or is set to gray if all weighst are equal to 1.
        Can either be used to save a file or open the connectome in your
        browser, depending on whether fp_out is None.

    :param idx_to_label: dict mapping each ROI index to its chord label
        (e.g., {0: "FPCN"})
    :param edges: list of tuples (a, b), where a = index of the 1st ROI &
        b = index of the 2nd ROI
    :param edge_weights: list of edge weights. The nth weight, here, should
        correspond to the nth edge in edges, above. If None or all the weights
        are equal to 1, then grey edges will be plotted.
    :param fp_glass: filepath where the connectome glass brain will be saved.
        If None, then the connectome will be loaded in your browser.
    :param coords: list of coordinates
    :param cmap: can either be None (default will be used), a string (will
        retrieve cmap from matplotlib), or a matplotlib colormap
    :param node_size: Size of the nodes (all must be constant)
    :param linewidths: linewidths used for the edges
    :param alphas: alphas used for the edges
    :param network_colors: dict mapping each label to a matplotlib color. If
        None, then default colors will be set based on matplotlib.
    :param network_order: If network_colors is None, then this list
        will specify the order in which default matplotlib colors will be
        assigned to networks.
    :param edge_threshold: This parameter acts the same as edge_threshold in
        nilearn.plotting.plot_connectome. Edges whose abs(weight) is under
        the threshold are omitted. edge_threshold can be a float or a string
        representing a percentile (e.g., "25"). The latter causes edges below
        the percentile to be omitted.
    """

    network_order, network_colors = \
        _network_order_colors_check(network_order, network_colors, idx_to_label)

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)

    if edge_weights is None:
        edge_weights = [1] * len(edges)
    else:
        edge_weights, edges = _threshold_proc(edge_threshold, edge_weights,
                                              edges)

    nonzero_node_idxs = list(set(node_idx for tup in edges for node_idx in tup))
    n = len(nonzero_node_idxs)
    adj = np.zeros((n, n))
    if all(weight == 1 for weight in edge_weights):
        node_idx_to_i = dict(
            (node_idx, i) for i, node_idx in enumerate(nonzero_node_idxs))
        for edge, edge_weight in zip(edges, edge_weights):
            adj[(node_idx_to_i[edge[0]], node_idx_to_i[edge[1]])] = 1
            adj[(node_idx_to_i[edge[1]], node_idx_to_i[edge[0]])] = 1
        if cmap is None: cmap = plt.get_cmap('Greys')
        if vmin is None: vmin = 0
        if vmin is None: vmax = 1
        edge_alphas = .4
    else:
        node_idx_to_i = dict(
            (node_idx, i) for i, node_idx in enumerate(nonzero_node_idxs))
        for edge, edge_weight in zip(edges, edge_weights):
            adj[(node_idx_to_i[edge[0]], node_idx_to_i[edge[1]])] = edge_weight
            adj[(node_idx_to_i[edge[1]], node_idx_to_i[edge[0]])] = edge_weight
        if cmap is None: cmap = plt.get_cmap('turbo')
        if vmin is None: vmin = min(edge_weights)
        if vmax is None: vmax = max(edge_weights)
        edge_alphas = .9

    node_coords_pruned = [coords[i] for i in nonzero_node_idxs]

    node_colors = [network_colors[idx_to_label[node_i]] for node_i in
                   nonzero_node_idxs]
    if isinstance(node_size, (list, np.ndarray)) and \
            (not len(node_size) == len(nonzero_node_idxs)):
        node_size_ = [node_size[i] for i in nonzero_node_idxs]
    else:
        node_size_ = node_size

    if fp_glass is None:
        cv = plotting.view_connectome(adj, node_coords_pruned, edge_cmap=cmap,
                                      node_color=node_colors,
                                      node_size=node_size_,
                                      linewidth=linewidths, colorbar=False)
        cv.open_in_browser()
    else:
        plotted_img = \
            plotting.plot_connectome(adj, node_coords_pruned,
                                     edge_cmap=cmap, node_color=node_colors,
                                     node_size=node_size_, edge_kwargs=
                                     {'linewidth': linewidths / 5,
                                      'alpha': edge_alphas if alphas is None
                                      else alphas},
                                     colorbar=False,
                                     edge_vmax=vmax, edge_vmin=vmin)
        plotted_img.savefig(fp_glass, dpi=dpi)
        plotted_img.close()



import os
from typing import Union

import numpy as np
from PIL import Image, ImageDraw, ImageFont
import matplotlib

from nichord import plot_chord, plot_glassbrain


def combine_imgs(fp_glass: str, fp_chord: str, fp_combined: str,
                 chord_mult: Union[float, int] = 1.05,
                 title: Union[None, str] = None,
                 only1glass: bool = False,
                 fontsize: int = 82) -> None:
    """
    Combine the glass brain connectome and chord diagram into a single picture.
        It works by loading the images with PIL then cropping and pasting
        them into a single image.

    :param fp_glass: filepath of the glass brain connectome
    :param fp_chord: filepath of the chord diagram
    :param fp_combined: filepath of the resulting image that combines the
        glass brain and chord diagrams
    :param chord_mult: specifies the size of the chord diagram relative to
        the glass brain. Probably shouldn't change.
    """

    image_glass = Image.open(fp_glass)
    image_chord = Image.open(fp_chord)

    w_glass, h_glass = image_glass.size
    w_chord, h_chord = image_chord.size

    w_chord_new = int(h_glass / h_chord * w_chord * chord_mult)
    h_chord_new = int(h_glass * chord_mult)
    image_chord = image_chord.resize((w_chord_new, h_chord_new))
    if only1glass: # drop coronal & axial
        image_glass = image_glass.crop((int(w_glass * .303), 0,
                                        int(w_glass * .70), h_glass))
    else: # drop only axial
        image_glass = image_glass.crop((int(w_glass * .303), 0, w_glass, h_glass))

    new_im = Image.new('RGB', (image_glass.size[0] + w_chord_new, int(
        (h_chord_new - (chord_mult - 1) / 2 * h_chord_new))),
                       color='white')
    new_im.paste(image_chord, (image_glass.size[0], 0))
    new_im.paste(image_glass, (0, int((chord_mult - 1) / 2 * h_chord_new)))

    new_fp = fp_combined

    if title is not None: new_im = add_title(title, new_im, fontsize=fontsize)
    new_im.save(new_fp)


def add_title(title: str, im, fontsize=82):
    font = ImageFont.truetype("arial.ttf", fontsize)
    im_w_title = Image.new(im.mode, (im.size[0], im.size[1] + 100), 'white')
    W, H = im.size[0], 100
    im_w_title.paste(im, (0, 100))
    draw = ImageDraw.Draw(im_w_title)
    _, _, w, h = draw.textbbox((0, 0), title, font=font)
    draw.text(((W-w)/2, (H-h)/2), title, font=font, fill='black', align='right')
    return im_w_title


def plot_and_combine(dir_out: str,
                     fn: str,
                     idx_to_label: dict,
                     edges: Union[list, np.ndarray],
                     edge_weights: Union[list, np.ndarray, None] = None,
                     coords: Union[list, np.ndarray] = None,
                     cmap: Union[None, str, matplotlib.colors.Colormap] = None,
                     network_order: Union[list, None] = None,
                     network_colors: Union[dict, None] = None,
                     chord_kwargs: Union[None, dict] = None,
                     glass_kwargs: Union[None, dict] = None,
                     title: Union[None, str] = None,
                     only1glass: bool = False,
                     fontsize: int = 82,
                     edge_threshold: Union[float, int, str] = 0.,
                     ) -> None:
    name, file_ext = os.path.splitext(fn)
    if file_ext == '':
        file_ext = '.png'

    dir_chord = os.path.join(dir_out, 'chord')
    if not os.path.isdir(dir_chord): os.mkdir(dir_chord)
    dir_glass = os.path.join(dir_out, 'glass')
    if not os.path.isdir(dir_glass): os.mkdir(dir_glass)

    fp_chord = os.path.join(dir_chord, f'{name}_chord{file_ext}')
    fp_glass = os.path.join(dir_glass, f'{name}_glass{file_ext}')
    fp_combined = os.path.join(dir_out, f'{name}{file_ext}')

    if chord_kwargs is None: chord_kwargs = {}
    if glass_kwargs is None: glass_kwargs = {}
    keys = ['network_order', 'network_colors', 'cmap', 'edge_threshold']
    vals = [network_order, network_colors, cmap, edge_threshold]
    for key, val in zip(keys, vals):
        if key not in chord_kwargs:
            chord_kwargs[key] = val
        if key not in glass_kwargs:
            glass_kwargs[key] = val

    plot_chord(idx_to_label, edges, edge_weights=edge_weights,
               coords=coords, fp_chord=fp_chord,
               **chord_kwargs)

    plot_glassbrain(idx_to_label, edges, edge_weights, fp_glass,
                    coords, **glass_kwargs)

    combine_imgs(fp_glass, fp_chord, fp_combined, title=title,
                 only1glass=only1glass, fontsize=fontsize)



import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.version_info[:3] < (3, 7, 0):
    print("Requires Python 3.7 to run.")
    sys.exit(1)

desc = '`NiChord` is a Python package for visualizing functional connectivity data. To find out more about the ' \
       'package and see a package, take a look at its [GitHub repo](https://github.com/paulcbogdan/NiChord). ' \
       'The package is also available via [conda](https://anaconda.org/conda-forge/nichord).'

setup(
    name="nichord",
    description="Creates chord diagrams for connectivity/graph data",
    long_description=desc,
    long_description_content_type="text/markdown",
    version="v0.3.0",
    packages=["nichord"],
    python_requires=">=3.7",
    url="https://github.com/paulcbogdan/NiChord",
    author="paulcbogdan",
    author_email="paulcbogdan@gmail.com",
    install_requires=["nilearn",
                      "pandas",
                      "matplotlib",
                      "numpy",
                      "scipy",
                      "pillow"],
    keywords=["plotting", "fmri", "plotting", "chord"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    license="MIT",
    license_file="LICENSE.txt"
)


import numpy as np
from typing import Union, Tuple


def convert_matrix(adj: Union[list, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Converts your adjacency (connectivity) matrix into a list of edges (i, j) 
        and their weights
    :param adj: the matrix
    """
    if isinstance(adj, list):
        adj = np.array(adj)
    idxs = np.tril_indices(adj.shape[0], k=-1)
    weights = adj[idxs]
    idxs = np.array(idxs).T
    smol = 1e-6
    idxs = idxs[(weights > smol) | (weights < -smol)]
    weights = weights[(weights > smol) | (weights < -smol)]
    return idxs, weights



'''
This code has been adapted from atlasreader: https://github.com/miykael/atlasreader
atlasreader has not been uploaded to conda-forge, and so parts of its code were
    copied here, so that NiChord can be uploaded to conda-forge.
'''

import numpy as np


def _read_atlas_peak(atlastype, coordinate):
    """
    Returns label of `coordinate` from corresponding `atlastype`

    If `atlastype` is probabilistic, `prob_thresh` determines (in percentage
    units) the threshold to apply before getting label of `coordinate`

    Parameters
    ----------
    atlastype : str
        Name of atlas to use
    coordinate : list of float
        x, y, z MNI coordinates of voxel
    prob_thresh : [0, 100] int, optional
        Probability (percentage) threshold to apply if `atlastype` is
        probabilistic

    Returns
    ------
    label : str or list of lists
        If `atlastype` is deterministic, this is the corresponding atlas label
        of `coordinate`. If `atlastype` is probabilistic, this is a list of
        lists where each entry denotes the probability and corresponding label
        of `coordinate`.
    """

    # get atlas data

    try:
        data = atlastype.image.get_fdata()
        # get voxel index
    except:
        data = atlastype.image

    voxID = _coord_xyz_to_ijk(atlastype.image.affine, coordinate).squeeze()
    voxID = _check_atlas_bounding_box(voxID, data.shape)
    labelID = int(data[voxID[0], voxID[1], voxID[2]])
    label = _get_label(atlastype, labelID)
    return label


def _coord_xyz_to_ijk(affine, coords):
    """
    Converts voxel `coords` in `affine` space to cartesian space

    Parameters
    ----------
    affine : (4, 4) array-like
        Affine matrix
    coords : (N,) list of list
        Image coordinate values, where each entry is a length three list of int
        denoting xyz coordinates in `affine` space

    Returns
    ------
    ijk : (N, 3) numpy.ndarray
        Provided `coords` in cartesian space
    """
    coords = _check_coord_inputs(coords)
    vox_coords = np.linalg.solve(affine, coords)[:3].T
    vox_coords = np.round(vox_coords).astype(int)
    return vox_coords


def _check_coord_inputs(coords):
    """
    Confirms `coords` are appropriate shape for coordinate transform

    Parameters
    ----------
    coords : array-like

    Returns
    -------
    coords : (4 x N) numpy.ndarray
    """
    coords = np.atleast_2d(coords).T
    if 3 not in coords.shape:
        raise ValueError('Input coordinates must be of shape (3 x N). '
                         'Provided coordinate shape: {}'.format(coords.shape))
    if coords.shape[0] != 3:
        coords = coords.T
    # add constant term to coords to make 4 x N
    coords = np.row_stack([coords, np.ones_like(coords[0])])
    return coords



def _get_label(atlastype, label_id):
    """
    Gets anatomical name of `label_id` in `atlastype`

    Parameters
    ----------
    atlastype : str
        Name of atlas to use
    label_id : int
        Numerical ID representing label

    Returns
    ------
    label : str
        Neuroanatomical region of `label_id` in `atlastype`
    """
    labels = atlastype.labels
    try:
        #print(atlastype)
        #print('test:', labels.query('index == {}'.format(label_id)).name.iloc[0])
        return labels.query('index == {}'.format(label_id)).name.iloc[0]
    except IndexError:
        return 'no_label'


def _check_atlas_bounding_box(voxIDs, box_shape):
    """
    Returns the provided voxel ID if the voxel is inside the bounding box of
    the atlas image, otherwise the voxel ID will be replaced with the origin.

    Parameters
    ----------
    voxIDs : (N, 3) numpy.ndarray
        `coords` in cartesian space
    box_shape : (3,) list of int
        size of the atlas bounding box

    Returns
    ------
    ijk : (N, 3) numpy.ndarray
        `coords` in cartesian space that are inside the bounding box
    """

    # Detect voxels that are outside the atlas bounding box
    vox_outside_box = np.sum(
        (voxIDs < 0) + (voxIDs >= box_shape[:3]), axis=-1, dtype='bool')

    # Set those voxels to the origin (i.e. a voxel outside the brain)
    voxIDs[vox_outside_box] = np.zeros(3, dtype='int')

    return voxIDs
